import json
import string
import slotter 

class SampleItemClass():
  def __init__(self, n, v):
    self.name = n
    self.value = v

def main():
  print "--------------------------"
  print "Create slotter"
  s = slotter.Slotter()
  print 'slotter_obj=%s' % s
  print "--------------------------"

  print "--------------------------"
  print "Create slots"
  s1 = s.add_slot(1, 10)
  print 's1.start=%d, s1.end=%d' % (s1.start, s1.end)
  s2 = s.add_slot(10, 20)
  print 's2.start=%d, s2.end=%d' % (s2.start, s2.end)
  print "--------------------------"

  print "--------------------------"
  print "Create items"
  # For strings - item.name == string name, item.value == string length
  item1 = s.create_item('item1')
  print 'item1.name=%s, item1.value=%d, item=%s' % (item1.name, item1.value, str(item1))
  item2 = s.create_item(string.ascii_lowercase[:15])
  print 'item2.name=%s, item2.value=%d, item=%s' % (item2.name, item2.value, str(item2))
  item3 = s.create_item(SampleItemClass('item3', 8))
  print 'item3.name=%s, item3.value=%d, item3.obj=%s' % (item3.name, item3.value, item3.obj)
  print "--------------------------"

  print "--------------------------"
  print "Add items to slots"
  s.add_item(item1)
  print 'item1.slot=%s' % item1.slot
  s.add_item(item2)
  print 'item2.slot=%s' % item2.slot
  s.add_item(item3)
  print 'item3.slot=%s' % item3.slot
  item4 = s.add_item('item4')
  print 'item4.name=%s, item4.value=%d, item=%s' % (item4.name, item4.value, str(item4))
  print 'item4.slot=%s' % item3.slot
  print 'item4.siblings=%s' % ([str(x) for x in item4.get_siblings()])
  item5 = s.add_item(11)
  print 'item5.name=%s, item5.value=%d, item=%s' % (item5.name, item5.value, str(item5))
  print 'item5.slot=%s' % item5.slot
  print 'item5.siblings=%s' % ([str(x) for x in item5.get_siblings()])
  print "--------------------------"

  print "--------------------------"
  print "Add an invalid item - i.e. one which doesn't fit in any slots"
  invalid_item = string.ascii_lowercase[:30]
  print 's.add_item(%s) = %s' % (invalid_item, s.add_item(invalid_item))
  print "--------------------------"

  print "--------------------------"
  print "Get all slots"
  print 'slots=%s' % s.slots
  print "--------------------------"

  print "--------------------------"
  print "Get all items"
  print 'items=%s' % s.items
  print "--------------------------"

  print "--------------------------"
  print "Get items of a particular slot"
  s1 = s.slots[0]
  print 's1.items=%s' % s1.items
  print "Get information about a specific item in that slot"
  i1 = s1.items[0]
  print "s1.first_item.name=%s, s1.first_item.value=%d" % (i1.name, i1.value)
  print "--------------------------"

  print "--------------------------"
  print "slot -> item mapping"
  print s.slot_items
  print "--------------------------"

  print "--------------------------"
  print "item -> slot mapping"
  print s.item_slots
  print "--------------------------"

  print "--------------------------"
  print "Dumpable slot -> item mapping"
  print json.dumps(s.dump(), indent=2)
  print "--------------------------"

  print "--------------------------"
  print "Dumpable item -> slot mapping"
  print json.dumps(s.dump(reverse=True), indent=2)
  print "--------------------------"

  print "--------------------------"
  print "Get slot of a particular item"
  s1 = s.get_slot(item1)
  print 'item1.slot.start=%d, item1.slot.end=%d' % (s1.start, s1.end)
  print "--------------------------"

  print "--------------------------"
  print "Remove an item object"
  print "before - s1.items=%s" % (s1.items)
  s.remove_item(item1)
  print "after - s1.items=%s" % (s1.items)
  print "Add item1 back"
  s.add_item(item1)
  print 'item1.slot=%s' % item1.slot
  print "--------------------------"

  print "--------------------------"
  print "Remove an item by name and value"
  print "before - s1.items=%s" % (s1.items)
  s.remove_item(name='item1', value=5)
  print "after - s1.items=%s" % (s1.items)
  print "--------------------------"

if __name__ == '__main__':
  main()



# -*- coding: utf-8 -*-
""" Shared utils for Slotter """

VALID_DERIVABLES = set([str, int, float])

class SlotterException(Exception):
  pass

def derive_item_value(item):
  """ For non user defined objects find their values """
  try:
    return getattr(item, 'value')
  except AttributeError as e:
    pass

  try:
    return getattr(item, 'get_item_value')()
  except AttributeError as e:
    pass

  for derivable_obj in VALID_DERIVABLES:
    if isinstance(item, derivable_obj):
      func = 'get_%s_value' % derivable_obj.__name__
      return globals()[func](item)


  raise SlotterException('Cannot derive item value - %s' % str(item))



# -*- coding: utf-8 -*-
""" Slot class """

from utils import SlotterException
from blist import sortedset

class Slot(object):
  """ That which contains items """
  def __init__(self, start, end, desc=None):
    if start > end or start == end:
      raise SlotterException('ERROR: Invalid start - %s, end - %s' % (start, end))

    self.start = start
    self.end = end
    self.items = sortedset(key=lambda x: x.value)

    if desc is None:
      desc = '%s-%s' % (str(self.start), str(self.end))

    self.desc = desc

  def __str__(self):
    return self.desc

  def can_accomodate(self, item):
    if item.value >= self.start and item.value < self.end:
      return True
    return False



# -*- coding: utf-8 -*-
""" Item class """

from utils import SlotterException

VALID_DERIVABLES = set([str, int, float])

def get_str_value(item):
  """ Get string value """
  return len(item)

def get_int_value(item):
  """ Get int value """
  return item

def get_float_value(item):
  """ Get float value """
  return item

class Item(object):
  """ Deafult item representation """
  def __init__(self, obj=None, name=None, value=None):
    """ Create the item to store """
    self.name = None
    self.value = None
    self.slot = None
    self.obj = self

    # try to derive the name and value of the item
    if obj is not None:
      self.obj = obj
      try:
        self.name = getattr(obj, 'name')
        self.value = getattr(obj, 'value')
      except AttributeError as e:
        pass

      try:
        self.name = getattr(obj, 'get_item_name')()
        self.value = getattr(obj, 'get_item_value')()
      except AttributeError as e:
        pass

      for derivable_obj in VALID_DERIVABLES:
        if isinstance(obj, derivable_obj):
          self.name = str(obj)
          self.value = globals()['get_%s_value' % derivable_obj.__name__](obj)

    if name is not None:
      self.name = name

    if value is not None:
      self.value = value

    failed_attrs = []
    for attr in ['name', 'value']:
      if getattr(self, attr) is None:
        failed_attrs.append(attr)

    if failed_attrs:
      failed_attrs_str = ','.join(failed_attrs)
      raise SlotterException('Invalid object - %s. Failed to find - %s' %\
                             (obj, failed_attrs_str))

  def __str__(self):
    """ Print this object as string """
    return '%s-%s' % (self.name, self.value)

  def get_siblings(self):
    """ Get siblings of this object """
    if self.slot is not None:
      all_items = set(self.slot.items)
      return list(all_items - set([self]))
    return []



# -*- coding: utf-8 -*-
""" The main class of slotter """

from blist import sortedset
from slot import Slot
from item import Item
import utils

class Slotter(object):
  """ Slotter class """

  def __init__(self, slots=None):
    """ Create the slots object """
    self.slots = sortedset(key=lambda x: x.start + x.end)
    if slots is not None:
      for slot in slots:
        self.slots.add(slot)
    self.verbose = False

    self.slot_items = {}
    self.item_slots = {}

    # prevent duplicate slots and items by remembering their string rep
    self._str_slots = {}
    self._str_items = {}

    self.items = sortedset(key = lambda x: x.value)

  def create_slot(self, start, end, desc=None):
    """ Just create a slot object - don't add it """
    return Slot(start, end, desc)

  def create_item(self, item):
    """ Just create a item object - don't slot it """
    return Item(item)

  def add_slot(self, start, end, desc=None):
    """ Define start, end for a slot """
    slot_obj = self.create_slot(start, end, desc)

    if str(slot_obj) in self._str_slots:
      return self._str_slots[str(slot_obj)]

    self.slots.add(slot_obj)
    self._str_slots[str(slot_obj)] = slot_obj

    return slot_obj

  def remove_slot(self, slot_obj):
    """ Define start, end for a slot """
    if slot_obj not in self.slots:
      return False
    for item in slot_obj.items:
      self.remove_item(item)
    self.slots.remove(slot_obj)
    del self._str_slots[str(slot_obj)]
    return True

  def _find_slot(self, item):
    """ Find the slot to which the item belongs """
    for slot in self.slots:
      if slot.can_accomodate(item):
        return slot
    return None

  def _find_item_by_name_value(self, name, value):
    """ Find item by its name and value """
    key = '%s-%s' % (str(name), str(value))
    if key in self._str_items:
      item = self._str_items[key]
      return item
    return None

  def add_item(self, item):
    """ Slot the item """
    if not isinstance(item, Item):
      item = self.create_item(item)

    if str(item) in self._str_items:
      return self._str_items[str(item)]

    slot = self._find_slot(item)
    if slot is None:
      if self.verbose:
        print 'ERROR: Did not find a slot - existing slots - %s' % self.dump().keys()
      return False

    if slot not in self.slot_items:
      self.slot_items[slot] = sortedset(key=lambda x: x.value)

    self.items.add(item)
    slot.items.add(item)
    self._str_items[str(item)] = item

    self.slot_items[slot].add(item)
    self.item_slots[item] = slot

    item.slot = slot

    return item

  def remove_item(self, item=None, name=None, value=None):
    """ Remove the slotted item """
    if item is None and name is not None and value is not None:
      item = self._find_item_by_name_value(name, value)

    if item is None or item not in self.item_slots:
      return False

    slot = self.item_slots[item]

    slot.items.remove(item)
    self.slot_items[slot].remove(item)
    del self._str_items[str(item)]

    self.items.remove(item)
    del self.item_slots[item]

    return True

  def dump(self, reverse=False):
    """ Dump the slotted ds """
    ds = {}
    if not reverse:
      for slot in self.slots:
        str_slot = str(slot)
        if str_slot not in ds:
          ds[str_slot] = []
        if slot in self.slot_items:
          items = self.slot_items[slot]
          for item in items:
            ds[str_slot].append((item.name, item.value))
      return ds

    for item in self.item_slots.keys():
      ds[str(item)] = str(self.item_slots[item])

    return ds

  def get_slot(self, item=None, name=None, value=None):
    """ Get slots """
    if item is None and name is not None and value is not None:
      item = self._find_item_by_name_value(name, value)
    if not isinstance(item, Item):
      item = self.create_item(item)
    if item is None:
      return None
    if str(item) in self._str_items:
      return self.item_slots[self._str_items[str(item)]]
    return None

  def get_items(self, slot=None, start=None, end=None):
    """ Get slot items """
    if slot is None and start is None and end is None:
      return list(self.items)

    if slot is not None:
      if slot in self.slot_items:
        return list(self.slot_items[slot])
    else:
      if start is not None and end is not None:
        key = '%s-%s' % (str(start), str(end))
        if key in self._str_slots:
          slot = self._str_slots[key]
          return list(self.slot_items[slot])

    return []
